/**
 * @author 陆益之
 * 此文件用于处理数据存储，数据存储共分为3个部分，库、文档、数据，每个库下面有多个文档，每个文档下面有多个数据，例如：
 * [users]
 * 		[lee]
 * 			[name = lee]
 * 			[age = 20]
 * 			[gender = male]
 * 			[love = [1,2]]
 * 		[jane]
 * 			[name = jane]
 * 			[age=19]
 * 			[gender = female]
 * 			[love = [2]]
 * [foods]
 * 		[1]
 * 			[id = 1]
 * 			[name = beef]
 * 		[2]
 * 			[id = 2]
 * 			[name = bread]
 */

/** 数据 */
export type IStorageDataItem = { [i: string]: any }

/** 文档（键=>值） */
export type IStorageDocument = { [i: string]: IStorageDataItem }

/** 库 */
export type IStorageDB = { [i: string]: IStorageDocument }


/** 数据存储 */
const storages: IStorageDB = {}

/**
 * 建立一个库操作器
 * @param db 库名称
 */
export function storage(db: string) {
	/**
	 * 是否存在数据库
	 */
	function exists(): boolean {
		return !!storages[db]
	}

	/** 
	 * 删除数据库 
	 */
	function drop() {
		delete storages[db]
	}

	/**
	 * 重命名数据库并返回新名称的数据库
	 * @param newName 新数据库名称
	 */
	function rename(newName: string) {
		storages[newName] = storages[db]
		drop()
		return storage(newName)
	}

	/** 
	 * 清空数据 
	 */
	function clean() {
		storages[db] = {}
	}

	/**
	 * 获取数据列表
	 * @param init 如果没有是否初始化
	 */
	function datas(init?: boolean): IStorageDocument | null {
		if (init && !storages[db]) storages[db] = {}
		return storages[db] || null
	}

	/**
	 * 通过键获取数据
	 * @param keys 键列表
	 */
	function keysof(keys: Array<string>) {
		const d = datas()
		if (!d) return null
		const buffer: Array<IStorageDataItem> = []
		keys.forEach(key => {
			if (d[key] === undefined) return
			buffer.push(d[key])
		})
		return buffer
	}

	/**
	 * 获取所有数据
	 */
	function all() {
		const d = datas()
		return d ? Object.values(d) : null
	}

	/**
	 * 数据查询
	 * @param option 查询选项
	 */
	function find(option: { [key: string]: string | boolean | number }) {
		const d = datas()
		if (!d) return null
		const okeys = Object.keys(option)
		//如果没有任何选项则返回所有数据
		if (!okeys.length) return all()
		//否则按照选项查询
		const buffer: Array<IStorageDataItem> = []
		for (let key in d) {
			let val = d[key]
			if (!val) continue
			let eq = true
			for (let okey in okeys) {
				if (option[okey] !== val[okey]) {
					eq = false
					break
				}
			}
			if (!eq) continue
			buffer.push(val)
		}
		return buffer
	}

	/**
	 * 删除数据
	 * @param keys 要删除的数据的键
	 */
	function _delete(keys: Array<string>) {
		const d = datas()
		if (!d) return
		keys.forEach(key => {
			delete d[key]
		})
	}

	/**
	 * 保存数据，无则添加
	 * @param key 要修改的数据的键
	 * @param value 数据值
	 * @param replace 是否进行替换操作，默认true
	 */
	function save(key: string, value: IStorageDataItem, replace: boolean = true) {
		const d = datas(true)!
		//没有数据或者说是替换则直接赋值
		if (d[key] || replace) {
			d[key] = value
			return
		}
		//按键覆盖
		const data = d[key]
		Object.keys(value).forEach(key => data[key] = value[key])
	}

	//返回操作函数
	return {
		exists, drop, rename, clean,
		datas, keysof, all, find,
		delete: _delete, save,
	}
}

/**
 * 获取所有库的键
 */
export function dbs() {
	return Object.keys(storages)
}